/*
   blocksys - block the system to artificially create a system latency

   Copyright (C) 2007,2018 Open Source Automation Development Lab (OSADL) eG
   Author: Carsten Emde <C.Emde@osadl.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/cpumask.h>
#include <linux/time.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/uaccess.h>

#include "version.h"

#define BLOCKSYS_MINOR MISC_DYNAMIC_MINOR

/*
 * These are the file operation function for user access to /dev/blocksys
 */
static ssize_t
blocksys_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
{
	char *mybuf;
	char which = 0;
	int nops;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	mybuf = kmalloc(count + 1, GFP_KERNEL);
	if (!mybuf)
		return -ENOMEM;

	if (copy_from_user(mybuf, buf, count)) {
		kfree(mybuf);
		return -EFAULT;
	}
	mybuf[count] = '\0';

	if (sscanf(mybuf, "%d%c", &nops, &which) >= 1) {
		struct timeval before, after, diff;
		int cpu = raw_smp_processor_id();
		char *label;

		switch (which) {
		case 'p': label = "preemption"; break;
		case 'i': label = "interrupts"; break;
		default: label = "preemption and interrupts";
			 which = 'b'; break;
		}
		printk(KERN_INFO "blocksys: %s of CPU #%d will be blocked "
		    "for %d nops\n", label, cpu, nops);
		do_gettimeofday(&before);
		if (which == 'p' || which == 'b')
			preempt_disable();
		if (which == 'i' || which == 'b')
			local_irq_disable();
		while (nops--)
			asm("nop");
		if (which == 'i' || which == 'b')
			local_irq_enable();
		if (which == 'p' || which == 'b')
			preempt_enable();
		do_gettimeofday(&after);
		diff.tv_sec = after.tv_sec - before.tv_sec;
		diff.tv_usec = after.tv_usec - before.tv_usec;
		if (diff.tv_usec < 0) {
			--diff.tv_sec;
			diff.tv_usec += 1000000;
		}
		printk(KERN_INFO "blocksys: %s of CPU #%d blocked "
		    "for about %ld us\n", label, cpu,
		    diff.tv_sec * 1000000 + diff.tv_usec);
	}
	kfree(mybuf);
	*ppos += count;
	return count;
}

static int
blocksys_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int
blocksys_release(struct inode *inode, struct file *file)
{
	return 0;
}

static struct file_operations blocksys_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.open		= blocksys_open,
	.write		= blocksys_write,
	.release	= blocksys_release,
};

static struct miscdevice blocksys_dev = {
	BLOCKSYS_MINOR,
	"blocksys",
	&blocksys_fops
};

static int __init blocksys_init(void)
{
	int ret;

	ret = misc_register(&blocksys_dev);
	if (ret)
		printk(KERN_ERR "blocksys: can't register dynamic misc device\n");
	else
		printk(KERN_INFO "blocksys driver v" VERSION " misc device %d\n",
			blocksys_dev.minor);
	return ret;
}

static void __exit blocksys_exit(void)
{
	misc_deregister(&blocksys_dev);
}

module_init(blocksys_init);
module_exit(blocksys_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Carsten Emde <C.Emde@osadl.org>");
MODULE_DESCRIPTION("Block the current CPU for a defined number of NOPs");
